"""Mock the fetching of file annotations from Deepen."""


#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#


def get_file_annotations(client_id, dataset_id, file_id):
    return {
        "labels": [
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:6",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [446, 215, 273, 211],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:7",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [263, 210, 235, 136],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:8",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [550, 201, 130, 98],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:9",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [424, 200, 109, 63],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:10",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [589, 185, 82, 70],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "adult",
                "label_id": "adult:15",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1046, 201, 154, 205],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "personal mobility",
                "label_id": "personal mobility:16",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [889, 230, 152, 130],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:18",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [321, 191, 85, 46],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "rigid",
                "label_id": "rigid:19",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [466, 129, 108, 112],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:20",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1050, 140, 230, 133],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:21",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [889, 154, 228, 80],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:22",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [852, 155, 211, 72],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:23",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [766, 162, 119, 48],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:24",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1100, 139, 164, 61],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:25",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1174, 139, 106, 60],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:26",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1244, 138, 36, 60],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:32",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [783, 162, 145, 49],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:35",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [1066, 149, 161, 49],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:36",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [918, 159, 101, 34],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
            {
                "client_id": client_id,
                "dataset_id": dataset_id,
                "file_id": file_id,
                "label_category_id": "car",
                "label_id": "car:37",
                "label_type": "box",
                "project_id": "2mjs8n87",
                "stage_id": "QA",
                "attributes": {},
                "box": [885, 162, 87, 29],
                "create_time_millis": 1568722395611,
                "labeller_email": "auto_interpolation",
                "update_time_millis": 1568722395611,
            },
        ]
    }
